//
// Example of using freeglut and glew to get OpenGL 3.3/4.1
//
// Author: Alexey V. Boreskov <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include	<GL/glew.h>

#ifdef	_WIN32
	#include	<GL/wglew.h>
#else
	#include	<GL/glxew.h>
#endif

#include	<freeglut.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#define	NUM_VERTICES	3
#define	VERTEX_SIZE		(3*sizeof(float))

GLuint 	vbo, vao;
GLuint	program;
float	scale = 0.3;

const char * vertexSource = 
"#version 330 core\n" \
"uniform float scale;\n"
"in vec3 position;\n" \
"void main(void)\n" \
"{\n" \
"	gl_Position   = vec4 ( scale * position, 1.0 );\n" \
"}\n";

const char * fragmentSource = 
"#version 330 core\n" \
"out vec4 color;\n"   \
"void main(void)\n"   \
"{\n"                 \
"	color = vec4 ( 1.0);\n" \
"}\n";


static const float vertices [] = 
{
   -2.0f, -2.0f, -1.0f, 
	0.0f,  2.0f, -1.0f, 
	2.0f, -2.0f, -1.0f
};

GLuint	loadShader ( const char * source, GLenum type )
{
    GLuint	shader = glCreateShader ( type );
    GLint   len    = strlen ( source );
    GLint   compileStatus;

    glShaderSource ( shader, 1, &source,  &len );

                                        // compile the shader
    glCompileShader ( shader );

    glGetShaderiv ( shader, GL_COMPILE_STATUS, &compileStatus );

	if ( compileStatus == 0 )			// some error
	{
		printf ( "Error comping shader.\n" );
		
		return 0;
	}

	return shader;
}

GLuint	createProgram ()
{
    GLuint	program = glCreateProgram ();
	GLuint	vs      = loadShader ( vertexSource, GL_VERTEX_SHADER );
	GLuint	fs      = loadShader ( fragmentSource, GL_FRAGMENT_SHADER );
	GLint	linked;
	
	glAttachShader ( program, vs );
	glAttachShader ( program, fs );

    glLinkProgram ( program );

    glGetProgramiv ( program, GL_LINK_STATUS, &linked );

    if ( !linked )
	{
		printf ( "Error linking program\n" );
		
        return 0;
	}
	
	return program;
}

bool	setAttrPtr ( GLuint program, const char * name, int numComponents, GLsizei stride, void * ptr, GLenum type = GL_FLOAT, bool normalized = false )
{
	int	loc = glGetAttribLocation ( program, name );

	if ( loc < 0 )
		return false;
		
	glVertexAttribPointer ( loc, 					// index
							numComponents, 			// number of values per vertex
							type, 					// type (GL_FLOAT)
							normalized ? GL_TRUE : GL_FALSE,
							stride, 				// stride (offset to next vertex data)
							(const GLvoid*) ptr );
		
	glEnableVertexAttribArray ( loc );
	
	return true;
}

bool    setUniformFloat ( GLuint program, const char * name, float value )
{
    int loc = glGetUniformLocation ( program, name );

    if ( loc < 0 )
        return false;

    glUniform1f ( loc, value );

    return true;
}

void init ()
{
	glClearColor ( 0.5, 0.5, 0.5, 1.0 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );
}

void display ()
{
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glUseProgram      ( program );
	glBindVertexArray ( vao );

	glDrawArrays      ( GL_TRIANGLES, 0, NUM_VERTICES );
	
	glBindVertexArray ( 0 );
    glUseProgram      ( 0 );

	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
	glViewport ( 0, 0, (GLsizei)w, (GLsizei)h );
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
}

void mouseWheel ( int wheel, int dir, int mouseX, int mouseY )
{
	if ( dir > 0 )
		scale += 0.05;
	else
		scale -= 0.05;
		
    glUseProgram      ( program );
	setUniformFloat   ( program, "scale", scale );
    glUseProgram      ( 0 );
	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 500, 500 );

								// prepare context for 3.3
	glutInitContextVersion ( 3, 3 );
	glutInitContextFlags   ( GLUT_FORWARD_COMPATIBLE | GLUT_DEBUG );
	glutInitContextProfile ( GLUT_CORE_PROFILE );

								// create window
	glutCreateWindow ( "GL 3.3 demo" );

	glewExperimental = GL_TRUE;

	glewInit ();

	if ( !GLEW_VERSION_3_3 )
	{
		printf ( "OpenGL 3.3 not supported.\n" );
		
		return 1;
	}

								// register handlers
	glutDisplayFunc    ( display );
	glutReshapeFunc    ( reshape );
	glutKeyboardFunc   ( key     );
	glutMouseWheelFunc ( mouseWheel );
	
	if ( !GL_ARB_vertex_array_object )
		printf ( "No VAO support\n" );

	program = createProgram ();
	
    glUseProgram ( program );

	glGenVertexArrays ( 1, &vao );
	glBindVertexArray ( vao );

	glGenBuffers ( 1, &vbo );
	glBindBuffer ( GL_ARRAY_BUFFER, vbo );
	glBufferData ( GL_ARRAY_BUFFER, NUM_VERTICES * VERTEX_SIZE, vertices, GL_STATIC_DRAW );	
	
	setAttrPtr      ( program, "position", 3, VERTEX_SIZE, (void *) 0 );
	setUniformFloat ( program, "scale", scale );

	glBindBuffer      ( GL_ARRAY_BUFFER, 0 );
	glBindVertexArray ( 0 );
	glUseProgram      ( 0 );

	glutMainLoop ();

	return 0;
}
